// Generated by CoffeeScript 1.4.0
var Connection, HttpConnection, TcpConnection, ar, aropt, bufferConcat, cursors, deconstructDatum, err, events, net, pb, r, util, varar,
  __hasProp = {}.hasOwnProperty,
  __extends = function(child, parent) { for (var key in parent) { if (__hasProp.call(parent, key)) child[key] = parent[key]; } function ctor() { this.constructor = child; } ctor.prototype = parent.prototype; child.prototype = new ctor(); child.__super__ = parent.prototype; return child; },
  __slice = [].slice;

net = require('net');

events = require('events');

util = require('./util');

err = require('./errors');

cursors = require('./cursor');

pb = require('./protobuf');

r = require('./ast');

ar = util.ar;

varar = util.varar;

aropt = util.aropt;

deconstructDatum = function(datum, opts) {
  var _this = this;
  return pb.DatumTypeSwitch(datum, {
    "R_NULL": function() {
      return null;
    },
    "R_BOOL": function() {
      return datum.r_bool;
    },
    "R_NUM": function() {
      return datum.r_num;
    },
    "R_STR": function() {
      return datum.r_str;
    },
    "R_ARRAY": function() {
      var dt, _i, _len, _ref, _results;
      _ref = datum.r_array;
      _results = [];
      for (_i = 0, _len = _ref.length; _i < _len; _i++) {
        dt = _ref[_i];
        _results.push(deconstructDatum(dt, opts));
      }
      return _results;
    },
    "R_OBJECT": function() {
      var obj, pair, _i, _len, _ref;
      obj = {};
      _ref = datum.r_object;
      for (_i = 0, _len = _ref.length; _i < _len; _i++) {
        pair = _ref[_i];
        obj[pair.key] = deconstructDatum(pair.val, opts);
      }
      switch (obj['$reql_type$']) {
        case 'TIME':
          switch (opts.timeFormat) {
            case 'native':
            case void 0:
              if (!(obj['epoch_time'] != null)) {
                throw new err.RqlDriverError("psudo-type TIME " + obj + " object missing expected field 'epoch_time'.");
              }
              return new Date(obj['epoch_time'] * 1000);
            case 'raw':
              return obj;
            default:
              throw new err.RqlDriverError("Unknown timeFormat run option " + opts.timeFormat + ".");
          }
          break;
        case void 0:
          return obj;
        default:
          throw new err.RqlDriverError("Unknown psudo-type " + obj['$reql_type$'] + ".");
      }
    }
  }, function() {
    throw new err.RqlDriverError("Unknown Datum type");
  });
};

Connection = (function(_super) {
  var mkAtom, mkErr, mkSeq;

  __extends(Connection, _super);

  Connection.prototype.DEFAULT_HOST = 'localhost';

  Connection.prototype.DEFAULT_PORT = 28015;

  Connection.prototype.DEFAULT_AUTH_KEY = '';

  Connection.prototype.DEFAULT_TIMEOUT = 20;

  function Connection(host, callback) {
    var conCallback, errCallback,
      _this = this;
    if (typeof host === 'undefined') {
      host = {};
    } else if (typeof host === 'string') {
      host = {
        host: host
      };
    }
    this.host = host.host || this.DEFAULT_HOST;
    this.port = host.port || this.DEFAULT_PORT;
    this.db = host.db;
    this.authKey = host.authKey || this.DEFAULT_AUTH_KEY;
    this.timeout = host.timeout || this.DEFAULT_TIMEOUT;
    this.outstandingCallbacks = {};
    this.nextToken = 1;
    this.open = false;
    this.buffer = new ArrayBuffer(0);
    this._events = this._events || {};
    errCallback = function(e) {
      _this.removeListener('connect', conCallback);
      if (e instanceof err.RqlDriverError) {
        return callback(e);
      } else {
        return callback(new err.RqlDriverError("Could not connect to " + _this.host + ":" + _this.port + "."));
      }
    };
    this.once('error', errCallback);
    conCallback = function() {
      _this.removeListener('error', errCallback);
      _this.open = true;
      return callback(null, _this);
    };
    this.once('connect', conCallback);
  }

  Connection.prototype._data = function(buf) {
    var response, responseBuffer, responseLength, _results;
    this.buffer = bufferConcat(this.buffer, buf);
    _results = [];
    while (this.buffer.byteLength >= 4) {
      responseLength = (new DataView(this.buffer)).getUint32(0, true);
      if (!(this.buffer.byteLength >= (4 + responseLength))) {
        break;
      }
      responseBuffer = this.buffer.slice(4, responseLength + 4);
      response = pb.ParseResponse(responseBuffer);
      this._processResponse(response);
      _results.push(this.buffer = this.buffer.slice(4 + responseLength));
    }
    return _results;
  };

  mkAtom = function(response, opts) {
    return deconstructDatum(response.response[0], opts);
  };

  mkSeq = function(response, opts) {
    var res, _i, _len, _ref, _results;
    _ref = response.response;
    _results = [];
    for (_i = 0, _len = _ref.length; _i < _len; _i++) {
      res = _ref[_i];
      _results.push(deconstructDatum(res, opts));
    }
    return _results;
  };

  mkErr = function(ErrClass, response, root) {
    var bt, frame, msg;
    msg = mkAtom(response);
    bt = (function() {
      var _i, _len, _ref, _results;
      _ref = response.backtrace.frames;
      _results = [];
      for (_i = 0, _len = _ref.length; _i < _len; _i++) {
        frame = _ref[_i];
        if (frame.type === "POS") {
          _results.push(parseInt(frame.pos));
        } else {
          _results.push(frame.opt);
        }
      }
      return _results;
    })();
    return new ErrClass(msg, root, bt);
  };

  Connection.prototype._delQuery = function(token) {
    delete this.outstandingCallbacks[token];
    if (Object.keys(this.outstandingCallbacks).length < 1 && !this.open) {
      return this.cancel();
    }
  };

  Connection.prototype._processResponse = function(response) {
    var cb, cursor, opts, root, token, _ref,
      _this = this;
    token = response.token;
    if (this.outstandingCallbacks[token] != null) {
      _ref = this.outstandingCallbacks[token], cb = _ref.cb, root = _ref.root, cursor = _ref.cursor, opts = _ref.opts;
      if (cursor != null) {
        return pb.ResponseTypeSwitch(response, {
          "SUCCESS_PARTIAL": function() {
            return cursor._addData(mkSeq(response, opts));
          },
          "SUCCESS_SEQUENCE": function() {
            cursor._endData(mkSeq(response, opts));
            return _this._delQuery(token);
          }
        }, function() {
          return cb(new err.RqlDriverError("Unknown response type"));
        });
      } else if (cb != null) {
        return pb.ResponseTypeSwitch(response, {
          "COMPILE_ERROR": function() {
            cb(mkErr(err.RqlCompileError, response, root));
            return _this._delQuery(token);
          },
          "CLIENT_ERROR": function() {
            cb(mkErr(err.RqlClientError, response, root));
            return _this._delQuery(token);
          },
          "RUNTIME_ERROR": function() {
            cb(mkErr(err.RqlRuntimeError, response, root));
            return _this._delQuery(token);
          },
          "SUCCESS_ATOM": function() {
            response = mkAtom(response, opts);
            if (Array.isArray(response)) {
              response = cursors.makeIterable(response);
            }
            cb(null, response);
            return _this._delQuery(token);
          },
          "SUCCESS_PARTIAL": function() {
            cursor = new cursors.Cursor(_this, token);
            _this.outstandingCallbacks[token].cursor = cursor;
            return cb(null, cursor._addData(mkSeq(response, opts)));
          },
          "SUCCESS_SEQUENCE": function() {
            cursor = new cursors.Cursor(_this, token);
            _this._delQuery(token);
            return cb(null, cursor._endData(mkSeq(response, opts)));
          }
        }, function() {
          return cb(new err.RqlDriverError("Unknown response type"));
        });
      }
    } else {
      return this.emit('error', new err.RqlDriverError("Unexpected token " + token + "."));
    }
  };

  Connection.prototype.close = ar(function() {
    return this.open = false;
  });

  Connection.prototype.cancel = ar(function() {
    return this.outstandingCallbacks = {};
  });

  Connection.prototype.reconnect = ar(function(callback) {
    var cb,
      _this = this;
    cb = function() {
      return _this.constructor.call(_this, {
        host: _this.host,
        port: _this.port
      }, callback);
    };
    return setTimeout(cb, 0);
  });

  Connection.prototype.use = ar(function(db) {
    return this.db = db;
  });

  Connection.prototype._start = function(term, cb, opts) {
    var pair, query, token;
    if (!this.open) {
      throw new err.RqlDriverError("Connection is closed.");
    }
    token = this.nextToken++;
    query = {
      'global_optargs': []
    };
    query.type = "START";
    query.query = term.build();
    query.token = token;
    if (this.db != null) {
      pair = {
        key: 'db',
        val: r.db(this.db).build()
      };
      query.global_optargs.push(pair);
    }
    if (opts.useOutdated != null) {
      pair = {
        key: 'use_outdated',
        val: r.expr(!!opts.useOutdated).build()
      };
      query.global_optargs.push(pair);
    }
    if (opts.noreply != null) {
      pair = {
        key: 'noreply',
        val: r.expr(!!opts.noreply).build()
      };
      query.global_optargs.push(pair);
    }
    if ((!(opts.noreply != null)) || !opts.noreply) {
      this.outstandingCallbacks[token] = {
        cb: cb,
        root: term,
        opts: opts
      };
    }
    this._sendQuery(query);
    if ((opts.noreply != null) && opts.noreply && typeof cb === 'function') {
      return cb(null);
    }
  };

  Connection.prototype._continueQuery = function(token) {
    var query;
    query = {
      type: "CONTINUE",
      token: token
    };
    return this._sendQuery(query);
  };

  Connection.prototype._endQuery = function(token) {
    var query,
      _this = this;
    query = {
      type: "STOP",
      token: token
    };
    this.outstandingCallbacks[token] = {
      cb: (function() {
        return _this._delQuery(token);
      }),
      root: null,
      opts: {}
    };
    return this._sendQuery(query);
  };

  Connection.prototype._sendQuery = function(query) {
    var data, finalArray, length;
    data = pb.SerializeQuery(query);
    length = data.byteLength;
    finalArray = new Uint8Array(length + 4);
    (new DataView(finalArray.buffer)).setInt32(0, length, true);
    finalArray.set(new Uint8Array(data), 4);
    return this.write(finalArray.buffer);
  };

  return Connection;

})(events.EventEmitter);

TcpConnection = (function(_super) {

  __extends(TcpConnection, _super);

  TcpConnection.isAvailable = function() {
    return !process.browser;
  };

  function TcpConnection(host, callback) {
    var timeout,
      _this = this;
    if (!TcpConnection.isAvailable()) {
      throw new err.RqlDriverError("TCP sockets are not available in this environment");
    }
    TcpConnection.__super__.constructor.call(this, host, callback);
    if (this.rawSocket != null) {
      this.close();
    }
    this.rawSocket = net.connect(this.port, this.host);
    this.rawSocket.setNoDelay();
    timeout = setTimeout((function() {
      _this.rawSocket.destroy();
      return _this.emit('error', new err.RqlDriverError("Handshake timedout"));
    }), this.timeout * 1000);
    this.rawSocket.once('error', function() {
      return clearTimeout(timeout);
    });
    this.rawSocket.once('connect', function() {
      var buf, buf_view, handshake_callback;
      buf = new ArrayBuffer(8);
      buf_view = new DataView(buf);
      buf_view.setUint32(0, 0x723081e1, true);
      buf_view.setUint32(4, _this.authKey.length, true);
      _this.write(buf);
      _this.rawSocket.write(_this.authKey, 'ascii');
      handshake_callback = function(buf) {
        var arr, b, i, status_buf, status_str, _i, _len, _ref;
        arr = util.toArrayBuffer(buf);
        _this.buffer = bufferConcat(_this.buffer, arr);
        _ref = new Uint8Array(_this.buffer);
        for (i = _i = 0, _len = _ref.length; _i < _len; i = ++_i) {
          b = _ref[i];
          if (b === 0) {
            _this.rawSocket.removeListener('data', handshake_callback);
            status_buf = _this.buffer.slice(0, i);
            _this.buffer = _this.buffer.slice(i + 1);
            status_str = String.fromCharCode.apply(null, new Uint8Array(status_buf));
            clearTimeout(timeout);
            if (status_str === "SUCCESS") {
              _this.rawSocket.on('data', function(buf) {
                return _this._data(util.toArrayBuffer(buf));
              });
              _this.emit('connect');
              return;
            } else {
              _this.emit('error', new err.RqlDriverError("Server dropped connection with message: \"" + status_str.trim() + "\""));
              return;
            }
          }
        }
      };
      return _this.rawSocket.on('data', handshake_callback);
    });
    this.rawSocket.on('error', function() {
      var args;
      args = 1 <= arguments.length ? __slice.call(arguments, 0) : [];
      return _this.emit.apply(_this, ['error'].concat(__slice.call(args)));
    });
    this.rawSocket.on('close', function() {
      _this.open = false;
      return _this.emit('close');
    });
  }

  TcpConnection.prototype.close = function() {
    TcpConnection.__super__.close.call(this);
    return this.rawSocket.end();
  };

  TcpConnection.prototype.cancel = function() {
    this.rawSocket.destroy();
    return TcpConnection.__super__.cancel.call(this);
  };

  TcpConnection.prototype.write = function(chunk) {
    var buf, byte, i, _i, _len, _ref;
    buf = new Buffer(chunk.byteLength);
    _ref = new Uint8Array(chunk);
    for (i = _i = 0, _len = _ref.length; _i < _len; i = ++_i) {
      byte = _ref[i];
      buf[i] = byte;
    }
    return this.rawSocket.write(buf);
  };

  return TcpConnection;

})(Connection);

HttpConnection = (function(_super) {

  __extends(HttpConnection, _super);

  HttpConnection.prototype.DEFAULT_PROTOCOL = 'http';

  HttpConnection.isAvailable = function() {
    return typeof XMLHttpRequest !== "undefined";
  };

  function HttpConnection(host, callback) {
    var protocol, url, xhr,
      _this = this;
    if (!HttpConnection.isAvailable()) {
      throw new err.RqlDriverError("XMLHttpRequest is not available in this environment");
    }
    HttpConnection.__super__.constructor.call(this, host, callback);
    protocol = host.protocol === 'https' ? 'https' : this.DEFAULT_PROTOCOL;
    url = "" + protocol + "://" + this.host + ":" + this.port + host.pathname + "ajax/reql/";
    xhr = new XMLHttpRequest;
    xhr.open("GET", url + "open-new-connection", true);
    xhr.responseType = "arraybuffer";
    xhr.onreadystatechange = function(e) {
      if (xhr.readyState === 4) {
        if (xhr.status === 200) {
          _this._url = url;
          _this._connId = (new DataView(xhr.response)).getInt32(0, true);
          return _this.emit('connect');
        } else {
          return _this.emit('error', new err.RqlDriverError("XHR error, http status " + xhr.status + "."));
        }
      }
    };
    xhr.send();
  }

  HttpConnection.prototype.cancel = function() {
    var xhr;
    xhr = new XMLHttpRequest;
    xhr.open("POST", "" + this._url + "close-connection?conn_id=" + this._connId, true);
    xhr.send();
    this._url = null;
    this._connId = null;
    return HttpConnection.__super__.cancel.call(this);
  };

  HttpConnection.prototype.write = function(chunk) {
    var xhr,
      _this = this;
    xhr = new XMLHttpRequest;
    xhr.open("POST", "" + this._url + "?conn_id=" + this._connId, true);
    xhr.responseType = "arraybuffer";
    xhr.onreadystatechange = function(e) {
      if (xhr.readyState === 4 && xhr.status === 200) {
        return _this._data(xhr.response);
      }
    };
    return xhr.send(chunk);
  };

  return HttpConnection;

})(Connection);

bufferConcat = function(buf1, buf2) {
  var view;
  view = new Uint8Array(buf1.byteLength + buf2.byteLength);
  view.set(new Uint8Array(buf1), 0);
  view.set(new Uint8Array(buf2), buf1.byteLength);
  return view.buffer;
};

module.exports.connect = ar(function(host, callback) {
  if (!(typeof host === 'string' || typeof host === 'object')) {
    throw new err.RqlDriverError("First argument to `connect` must be a string giving the " + "host to `connect` to or an object giving `host` and `port`.");
  }
  if (typeof callback !== 'function') {
    throw new err.RqlDriverError("Second argument to `connect` must be a callback to invoke with " + "either an error or the successfully established connection.");
  }
  if (TcpConnection.isAvailable()) {
    new TcpConnection(host, callback);
  } else if (HttpConnection.isAvailable()) {
    new HttpConnection(host, callback);
  } else {
    throw new err.RqlDriverError("Neither TCP nor HTTP avaiable in this environment");
  }
});
